/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

#include <../config.h>

#ifdef ENABLE_ROS

#include <ros/ros.h>
#include "sensor_msgs/Image.h"
#include "sensor_msgs/fill_image.h"
#include "fdopencv_msgs/ImageMap.h"
#include "fdopencv_msgs/ImageMapVal.h"
#include "HarkRosGlobals.h"

using namespace std;
using namespace FD;
  
class CvfdhRosImagePublisher;
  
DECLARE_NODE(CvfdhRosImagePublisher)
/*Node
 *
 * @name CvfdhRosImagePublisher
 * @category FDOPENCV_HARK:ImageIO
 * @description Publish an image to ROS
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 * 
 * @parameter_name TOPIC_NAME_IMAGE
 * @parameter_type string
 * @parameter_value Image
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name TOPIC_NAME_IMAGEMAP
 * @parameter_type string
 * @parameter_value ImageMap
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name BUFFER_NUM
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description Buffer size for a ROS published message
 * 
 * @parameter_name ROS_LOOP_RATE
 * @parameter_type float
 * @parameter_value 100000
 * @parameter_description This allows you to specify a frequency that you would like to loop at [Hz]. Keep this value large. (If ROS interval is shorter than HARK interval, ROS interval is overwritten.)
 *
 END*/
  
  
class CvfdhRosImagePublisher : public BufferedNode {
    
  int m_imageInID;
  int m_srcInID;
  int m_featOutID;

  // ROS related parameters
  string topic_name_image;
  string topic_name_imagemap;
  int buffer_num;
  float ros_loop_rate;

private:
  ros::Publisher _pub_Image;
  ros::Publisher _pub_ImageMap;
    
public:
  CvfdhRosImagePublisher(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {

    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    topic_name_image     = object_cast<String>(parameters.get("TOPIC_NAME_IMAGE"));
    topic_name_imagemap  = object_cast<String>(parameters.get("TOPIC_NAME_IMAGEMAP"));
    buffer_num  = dereference_cast<int>(parameters.get("BUFFER_NUM"));
    ros_loop_rate  = dereference_cast<float>(parameters.get("ROS_LOOP_RATE"));
    inOrder = true;
    cout << getName() << " constructor end..." << endl;
  }
      
  ~CvfdhRosImagePublisher()
  {
  }

  virtual void initialize()
  {
    cout << getName() << " initialized..." << endl;
    _pub_Image    = __MasterRosNodeHandler__->advertise<sensor_msgs::Image>(topic_name_image, buffer_num, false);
    _pub_ImageMap = __MasterRosNodeHandler__->advertise<fdopencv_msgs::ImageMap>(topic_name_imagemap, buffer_num, false);
    this->BufferedNode::initialize();
  }

  virtual int translateInput(string inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    // ROS loop settings
    ros::Rate loop_rate(ros_loop_rate);

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);

    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;

      // Publishing Image message
      sensor_msgs::Image _img;
      _img.header.frame_id = "";
      _img.header.stamp = ros::Time::now();
      fillImage(_img, "rgb8", img_in_tmp->height, img_in_tmp->width, 3 * img_in_tmp->width, img_in_tmp->imageData);
      _pub_Image.publish(_img);

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }

    }

    IplImage* img_in;      
    fdopencv_msgs::ImageMap _imgmap;
    _imgmap.header.frame_id = "";
    _imgmap.header.stamp = ros::Time::now();
    _imgmap.count = count;
    int mapnum = 0;

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      img_in = cvCloneImage(imagePtr.getImage());
      // Make the published message
      fdopencv_msgs::ImageMapVal _imgmapval;
      _imgmapval.id = img_it->first;
      fillImage(_imgmapval.img, "rgb8", img_in->height, img_in->width, 3 * img_in->width, img_in->imageData);
      _imgmap.src.push_back(_imgmapval);
      mapnum++;
    }
    
    _imgmap.num = mapnum;

    _pub_ImageMap.publish(_imgmap);
      
    cvReleaseImage (&img_in);
    cvReleaseImage(&img_in_tmp);

    loop_rate.sleep();

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhRosImagePublisher)
};

#endif
