/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include <math.h>
#include "SrcInfo.h"

#include <boost/scoped_ptr.hpp>
#include <boost/lexical_cast.hpp>

using namespace std;
using namespace boost;
using namespace FD;

class CvfdhSrcInfoTracker;

DECLARE_NODE(CvfdhSrcInfoTracker);
/*Node
 *
 * @name CvfdhSrcInfoTracker
 * @category FDOPENCV_HARK:ROI
 * @description CvfdhSrcInfoTracker
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source segmentation results..
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @input_description Source segmentation with ID. 
 *
 * @parameter_name USE_ROI_SIZE
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Use ROI selection with respect to ROI_SIZE.
 *
 * @parameter_name ROI_SIZE
 * @parameter_type object
 * @parameter_value <Vector<int> 0 0>
 * @parameter_valid USE_ROI_SIZE=true
 * @parameter_description Enable elimination of ROIs with respect to ROI size (width * height). This is a 2D vector representing [maximum minimum] size of the ROI selection. Use 0 to disable one.
 *
 * @parameter_name USE_ASPECT_RATIO
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Use ROI selection with respect to ASPECT_RATIO.
 *
 * @parameter_name ASPECT_RATIO
 * @parameter_type object
 * @parameter_value <Vector<float> 0.0 0.0>
 * @parameter_valid USE_ASPECT_RATIO=true
 * @parameter_description Enable elimination of ROIs with respect to ROI aspect ratio (height / width). This is a 2D vector representing [maximum minimum] ratio of ROIs. Use 0 to disable one.
 *
 * @parameter_name USE_PAUSE_LENGTH
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Use ROI selection with respect to PAUSE_LENGTH.
 *
 * @parameter_name PAUSE_LENGTH
 * @parameter_type float
 * @parameter_value 100.0
 * @parameter_valid USE_PAUSE_LENGTH=true
 * @parameter_description Life duration of source in frame num. 
 *
 * @parameter_name USE_MIN_SRC_INTERVAL
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Use ROI selection with respect to MIN_SRC_INTERVAL.
 * 
 * @parameter_name MIN_SRC_INTERVAL
 * @parameter_type float
 * @parameter_value 10.0
 * @parameter_valid USE_MIN_SRC_INTERVAL=true
 * @parameter_description Source euclid diatance interval threshold in pixels. 
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Output debug information if true
 *
 END*/

class CvfdhSrcInfoTracker : public BufferedNode {
  int srcinID;
  int srcoutID;

  bool use_roi_size;
  Vector<int> roi_size;
  bool use_aspect_ratio;
  Vector<float> aspect_ratio;
  bool use_pause_length;
  float pause_length;
  bool use_min_src_interval;
  float min_src_interval;
  bool enable_debug;

  int next_id;
  RCPtr<Vector<ObjectRef> > plast_src;
  RCPtr<Vector<ObjectRef> > pnew_src;

public:
  CvfdhSrcInfoTracker(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    srcinID = addInput("SRCIN");
    srcoutID = addOutput("SRCOUT");

    cout << "deb 1" << endl;

    use_roi_size = dereference_cast<bool>(parameters.get("USE_ROI_SIZE"));
    roi_size = object_cast<Vector<int> >(parameters.get("ROI_SIZE"));

    use_aspect_ratio = dereference_cast<bool>(parameters.get("USE_ASPECT_RATIO"));
    aspect_ratio = object_cast<Vector<float> >(parameters.get("ASPECT_RATIO"));

    cout << "deb 1" << endl;

    use_pause_length = dereference_cast<bool>(parameters.get("USE_PAUSE_LENGTH"));
    if(use_pause_length)
      pause_length = dereference_cast<float>(parameters.get("PAUSE_LENGTH"));
    else
      pause_length = 1.0;

    cout << "deb 1" << endl;

    use_min_src_interval = dereference_cast<bool>(parameters.get("USE_MIN_SRC_INTERVAL"));
    if(use_min_src_interval)
      min_src_interval = dereference_cast<float>(parameters.get("MIN_SRC_INTERVAL"));
    else
      min_src_interval = 0.0;

    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
    next_id = 0;    
    plast_src = Vector<ObjectRef>::alloc(0);

    inOrder = true;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Vector<ObjectRef> > pcur_src = getInput(srcinID, count);
    const Vector<ObjectRef>& cur_sources = *pcur_src;

    pnew_src = Vector<ObjectRef>::alloc(0);
    Vector<ObjectRef>& new_sources = *pnew_src;
    Vector<ObjectRef>& last_sources = *plast_src;
    Vector<ObjectRef> remaining_sources(cur_sources);
    
    // Process for source for the previous sources (for eliminating)
    for (int i = 0; i < last_sources.size(); i++) {
      RCPtr<SrcInfo> last_src = last_sources[i];

      // Life end
      if (last_src->remaining < 0) {
        cerr << "Source " << last_src->id << " is removed.\n";
        continue;
      }

      // Time Association Candidate
      float min_dist = 10000.0;
      int min_id = 0;
      for (int j = 0; j < remaining_sources.size(); j++) {
        RCPtr<SrcInfo> cur_src = remaining_sources[j];
                
	if((cur_src->roi_xy[0] * cur_src->roi_xy[1] < roi_size[0])&&(use_roi_size)&&(roi_size[0] != 0)){
          continue;
	}
	if((cur_src->roi_xy[0] * cur_src->roi_xy[1] > roi_size[1])&&(use_roi_size)&&(roi_size[1] != 0)){
          continue;
	}
	if(((float)cur_src->roi_xy[1] / (float)cur_src->roi_xy[0] < aspect_ratio[0])&&(use_aspect_ratio)&&(aspect_ratio[0] != 0.0)){
          continue;
	}
	if(((float)cur_src->roi_xy[1] / (float)cur_src->roi_xy[0] > aspect_ratio[1])&&(use_aspect_ratio)&&(aspect_ratio[1] != 0.0)){
          continue;
	}
	
	float tmp_xx = (float)abs(cur_src->roi_c[0] - last_src->roi_c[0]);
	float tmp_yy = (float)abs(cur_src->roi_c[1] - last_src->roi_c[1]);
	float tmp_dist = sqrt(tmp_xx * tmp_xx + tmp_yy * tmp_yy);	
        if (tmp_dist < min_dist) {
          min_dist = tmp_dist;
          min_id = j;
        }
      }
      
      // Time Association
      if (min_dist < min_src_interval) {
        RCPtr<SrcInfo> old = remaining_sources[min_id]->clone();
        old->remaining = pause_length;
        old->id = last_src->id;
        new_sources.push_back(old);
        remaining_sources.erase(remaining_sources.begin() + min_id);
      }
      else {
        RCPtr<SrcInfo> old = last_src->clone();
        old->remaining -= 1;
        new_sources.push_back(old);
      }
    }

    // Process for source for the current sources (for generating)
    for (int i = 0; i < cur_sources.size(); i++) {
      RCPtr<SrcInfo> cur_src = cur_sources[i];
            
      // Elimination
      if((cur_src->roi_xy[0] * cur_src->roi_xy[1] < roi_size[0])&&(use_roi_size)&&(roi_size[0] != 0)){
	continue;
      }
      if((cur_src->roi_xy[0] * cur_src->roi_xy[1] > roi_size[1])&&(use_roi_size)&&(roi_size[1] != 0)){
	continue;
      }
      if(((float)cur_src->roi_xy[1] / (float)cur_src->roi_xy[0] < aspect_ratio[0])&&(use_aspect_ratio)&&(aspect_ratio[0] != 0)){
	continue;
      }
      if(((float)cur_src->roi_xy[1] / (float)cur_src->roi_xy[0] > aspect_ratio[1])&&(use_aspect_ratio)&&(aspect_ratio[1] != 0)){
	continue;
      }

      float min_dist = 10000.0;
      int min_id = 0;
      for (int j = 0; j < new_sources.size(); j++) {
        RCPtr<SrcInfo> new_src = new_sources[j];
                
	float tmp_xx = (float)abs(cur_src->roi_c[0] - new_src->roi_c[0]);
	float tmp_yy = (float)abs(cur_src->roi_c[1] - new_src->roi_c[1]);
	float tmp_dist = sqrt(tmp_xx * tmp_xx + tmp_yy * tmp_yy);	
        if (tmp_dist < min_dist) {
          min_dist = tmp_dist;
          min_id = j;
        }
      }
	    
      if (min_dist > min_src_interval) {
        cerr << "Source " << next_id << " is created.\n";
        cur_src->id = next_id++;
        cur_src->remaining = pause_length;
        new_sources.push_back(cur_sources[i]);
      }
    }

    if (enable_debug == true) {
	fprintf(stdout, "time: %d, ", count);
	for (int i = 0; i < new_sources.size(); i++) {
	    RCPtr<SrcInfo> src = new_sources[i];
	    
	    fprintf(stdout, "ID: %d, roi c : %e %e, roi xy %e %e, ",
		    src->id,
		    src->roi_c[0],
		    src->roi_c[1],
		    src->roi_xy[0],
		    src->roi_xy[1]);
	}
	fprintf(stdout, "\n");
    }


    plast_src = pnew_src;
    out[count] = pnew_src;
  }
};
